unit main_test_curves;
{$I G32i.inc}

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs, G32_Interface,
  StdCtrls, ExtCtrls, GR32_Transforms, Buttons, ComCtrls;

type
  TfmMain = class(TForm)
    Panel1: TPanel;
    sbMain: TStatusBar;
    btnPolyBezier: TButton;
    btnEllipse: TButton;
    GroupBox1: TGroupBox;
    cbAntialising: TCheckBox;
    cbFilling: TCheckBox;
    btnDrawRotatedEllipsed: TButton;
    udAngle1: TUpDown;
    edAngle1: TEdit;
    edAngle2: TEdit;
    udAngle2: TUpDown;
    btnClearBitmap: TButton;
    Label1: TLabel;
    Label2: TLabel;
    GroupBox2: TGroupBox;
    cbEllipticFigures: TCheckBox;
    btnDrawSegments: TButton;
    btnDrawPies: TButton;
    btnDrawArc: TButton;
    btnDrawRoundedRectangles: TButton;
    updRoundRadius: TUpDown;
    edRoundRadius: TEdit;
    Label3: TLabel;
    btnDrawTransformedEllipse: TButton;
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure FormResize(Sender: TObject);

    procedure btnPolyBezierClick(Sender: TObject);
    procedure btnEllipseClick(Sender: TObject);
    procedure cbAntialisingClick(Sender: TObject);
    procedure cbFillingClick(Sender: TObject);
     procedure btnDrawRotatedEllipsedClick(Sender: TObject);
    procedure btnDrawArcClick(Sender: TObject);
    procedure btnClearBitmapClick(Sender: TObject);
    procedure FormPaint(Sender: TObject);
    procedure btnDrawSegmentsClick(Sender: TObject);
    procedure btnDrawPiesClick(Sender: TObject);
    procedure btnDrawRoundedRectanglesClick(Sender: TObject);
    procedure btnDrawTransformedEllipseClick(Sender: TObject);
  private
    { Private declarations }
    fBuffer : tBitmap32Ex;
    fDrawningOptions : tPolygonDrawOptions;
  public
    { Public declarations }
    procedure UpdateOptions;
    procedure RepaintBuffer;
  end;

var
  fmMain: TfmMain;



implementation

uses GR32, GR32_Polygons, Math
{$IFDEF DEBUG_OPTIMIZATION}
,Debug_LE
{$ENDIF}
;

{$R *.DFM}

procedure TfmMain.FormCreate(Sender: TObject);
begin
  Self.Color := clBlack;
  fBuffer := tBitmap32Ex.Create;

  UpdateOptions;

  Application.HintPause := 100;

  fBuffer.SetSize(Self.ClientWidth, Self.ClientHeight);
end;

procedure TfmMain.FormDestroy(Sender: TObject);
begin
  fBuffer.Free;
end;

procedure TfmMain.FormResize(Sender: TObject);
begin
  fBuffer.SetSize(Self.ClientWidth, Self.ClientHeight);
end;

procedure TfmMain.btnPolyBezierClick(Sender: TObject);
const
{$IFDEF DEBUG_OPTIMIZATION}
  N = 10000;
{$ELSE}
  N = 100;
{$ENDIF}
var
  PP : tArrayOfFixedPoint;
  i : integer;
begin
  {     3*N + 1, N -   }
  {$IFDEF DEBUG_OPTIMIZATION}  Secundomer.Start; {$ENDIF DEBUG_OPTIMIZATION}
  SetLength(PP, 3*N+1);
  PP[0].x := Fixed(random(ClientWidth)); PP[0].y := Fixed(random(ClientWidth));
  for i := 0 to N - 1 do
    begin
    PP[3*i + 1].X := Fixed(random(ClientWidth)); PP[3*i + 1].y := Fixed(random(ClientWidth));
    PP[3*i + 2].X := Fixed(random(ClientWidth)); PP[3*i + 2].y := Fixed(random(ClientWidth));
    PP[3*i + 3].X := Fixed(random(ClientWidth)); PP[3*i + 3].y := Fixed(random(ClientWidth));
    end;

  fBuffer.PolyBezier(PP, clTrRed32 or $C0, fDrawningOptions, true);
  SetLength(PP, 0);
  {$IFDEF DEBUG_OPTIMIZATION}  Secundomer.Stop; {$ENDIF DEBUG_OPTIMIZATION}
  RepaintBuffer;
  {$IFDEF DEBUG_OPTIMIZATION}  Secundomer.ShowResult; {$ENDIF DEBUG_OPTIMIZATION}
end;

procedure TfmMain.btnEllipseClick(Sender: TObject);
const
{$IFDEF DEBUG_OPTIMIZATION}
  N = 10000;
{$ELSE}
  N = 100;
{$ENDIF}
var
  i : integer;
  vRect : tFixedRect;
  x, y : integer;
begin
  {$IFDEF DEBUG_OPTIMIZATION}  Secundomer.Start; {$ENDIF DEBUG_OPTIMIZATION}
  for i := 1 to N  do
    begin
    x := random(ClientWidth);
    y := random(ClientHeight);
    vRect := FixedRect(rect(x, y, x + random(100), y + random(100)));
    fBuffer.Ellipse(vRect , clTrBlue32, fDrawningOptions);
    end;
  {$IFDEF DEBUG_OPTIMIZATION}  Secundomer.Stop; {$ENDIF DEBUG_OPTIMIZATION}
  RepaintBuffer;
  {$IFDEF DEBUG_OPTIMIZATION}  Secundomer.ShowResult; {$ENDIF DEBUG_OPTIMIZATION}
end;

procedure TfmMain.btnDrawRotatedEllipsedClick(Sender: TObject);
const
{$IFDEF DEBUG_OPTIMIZATION}
  N = 10000;
{$ELSE}
  N = 100;
{$ENDIF}
var
  i : integer;
  vA, vB : GR32.TFixed;
  vC : GR32.TFixedPoint;
  vAngle : double;
begin
  {$IFDEF DEBUG_OPTIMIZATION}  Secundomer.Start; {$ENDIF DEBUG_OPTIMIZATION}
  for i := 1 to N  do
    begin
    vA := GR32.Fixed(random(100));
    vB := GR32.Fixed(random(100));
    vC.x := GR32.Fixed(random(ClientWidth)); vC.y := GR32.Fixed(random(ClientWidth));
    vAngle := DegToRad(udAngle1.Position);

    fBuffer.EllipseRotated(vC, vA, vB, vAngle, clTrGreen32, fDrawningOptions);
    end;
  {$IFDEF DEBUG_OPTIMIZATION}  Secundomer.Stop; {$ENDIF DEBUG_OPTIMIZATION}
  RepaintBuffer;
  {$IFDEF DEBUG_OPTIMIZATION}  Secundomer.ShowResult; {$ENDIF DEBUG_OPTIMIZATION}
end;

procedure TfmMain.btnDrawTransformedEllipseClick(Sender: TObject);
const
{$IFDEF DEBUG_OPTIMIZATION}
  N = 10000;
{$ELSE}
  N = 100;
{$ENDIF}
var
  i : integer;
  vA, vB : GR32.TFixed;
  vC : GR32.TFixedPoint;
  vAT : TAffineTransformation;
begin
  {$IFDEF DEBUG_OPTIMIZATION}  Secundomer.Start; {$ENDIF DEBUG_OPTIMIZATION}
  vAT := TAffineTransformation.Create;                                
  vAT.Rotate(0, 0, udAngle1.Position);
  for i := 1 to N  do
    begin
    vA := GR32.Fixed(random(100));
    vB := GR32.Fixed(random(100));
    vC.x := GR32.Fixed(random(ClientWidth)); vC.y := GR32.Fixed(random(ClientWidth));
    gEllipseT(fBuffer, vC, vA, vB, vAT.Matrix, clTrGreen32, fDrawningOptions);
    end;
  {$IFDEF DEBUG_OPTIMIZATION}  Secundomer.Stop; {$ENDIF DEBUG_OPTIMIZATION}
  RepaintBuffer;
  {$IFDEF DEBUG_OPTIMIZATION}  Secundomer.ShowResult; {$ENDIF DEBUG_OPTIMIZATION}
end;

procedure TfmMain.btnDrawArcClick(Sender: TObject);
var
  i : integer;
  vA, vB : GR32.TFixed;
  vC : GR32.TFixedPoint;
  vAngle1 : double;
  vAngle2 : double;
begin
  for i := 1 to 100  do
    begin
    vC.x := GR32.Fixed(random(ClientWidth)); vC.y := GR32.Fixed(random(ClientWidth));
    vAngle1 := DegToRad(udAngle1.Position);
    vAngle2 := DegToRad(udAngle2.Position);

    vA := GR32.Fixed(random(100));
    if cbEllipticFigures.Checked then
      begin
      vB := GR32.Fixed(random(100));
      fBuffer.ArcElliptic(vC, vA, vB, vAngle1, vAngle2, clTrGreen32, fDrawningOptions);
      end
    else
      begin
      fBuffer.Arc(vC, vA, vAngle1, vAngle2, clTrGreen32, fDrawningOptions);
      end;
    end;
  RepaintBuffer;
end;

procedure TfmMain.UpdateOptions;
begin
  fDrawningOptions := 0;
  if cbAntialising.Checked then fDrawningOptions := pdoFloat;
  if cbFilling.Checked then fDrawningOptions := fDrawningOptions or pdoFilling;
end;

procedure TfmMain.cbAntialisingClick(Sender: TObject);
begin
  UpdateOptions;
end;


procedure TfmMain.cbFillingClick(Sender: TObject);
begin
  UpdateOptions;
end;

procedure TfmMain.btnClearBitmapClick(Sender: TObject);
begin
  fBuffer.FillRect(0, 0, fBuffer.Width, fBuffer.Height, clBlack32);
  RepaintBuffer;
end;

procedure TfmMain.RepaintBuffer;
begin
  fBuffer.DrawTo(Canvas.Handle, 0, 0);
end;

procedure TfmMain.FormPaint(Sender: TObject);
begin
  RepaintBuffer;
end;

procedure TfmMain.btnDrawPiesClick(Sender: TObject);
var
  i : integer;
  vA, vB : GR32.TFixed;
  vC : GR32.TFixedPoint;
  vAngle1 : double;
  vAngle2 : double;
begin
  for i := 1 to 100  do
    begin
    vC.x := GR32.Fixed(random(ClientWidth)); vC.y := GR32.Fixed(random(ClientWidth));
    vAngle1 := DegToRad(udAngle1.Position);
    vAngle2 := DegToRad(udAngle2.Position);

    vA := GR32.Fixed(random(100));
    if cbEllipticFigures.Checked then
      begin
      vB := GR32.Fixed(random(100));
      fBuffer.PieElliptic(vC, vA, vB, vAngle1, vAngle2, clTrRed32, fDrawningOptions);
      end
    else
      begin
      fBuffer.Pie(vC, vA, vAngle1, vAngle2, clTrRed32, fDrawningOptions);
      end;
    end;
  RepaintBuffer;
end;


procedure TfmMain.btnDrawSegmentsClick(Sender: TObject);
var
  i : integer;
  vA, vB : GR32.TFixed;
  vC : GR32.TFixedPoint;
  vAngle1 : double;
  vAngle2 : double;
begin
  for i := 1 to 100  do
    begin
    vC.x := GR32.Fixed(random(ClientWidth)); vC.y := GR32.Fixed(random(ClientWidth));
    vAngle1 := DegToRad(udAngle1.Position);
    vAngle2 := DegToRad(udAngle2.Position);

    vA := GR32.Fixed(random(100));
    if cbEllipticFigures.Checked then
      begin
      vB := GR32.Fixed(random(100));
      fBuffer.SegmentElliptic(vC, vA, vB, vAngle1, vAngle2, clTrBlue32, fDrawningOptions);
      end
    else
      begin
      fBuffer.Segment(vC, vA, vAngle1, vAngle2,clTrBlue32, fDrawningOptions);
      end;
    end;
  RepaintBuffer;
end;


procedure TfmMain.btnDrawRoundedRectanglesClick(Sender: TObject);
var
  i : integer;
  vRect : tFixedRect;
  x, y : integer;
  xR : integer;
  w, h : integer;
  vA   : double;
const
{$IFDEF DEBUG_OPTIMIZATION}
   N = 1000;
{$ELSE}
   N = 100;
{$ENDIF}
begin
  xR := updRoundRadius.Position;
  vA := DegToRad(udAngle1.Position);
  {$IFDEF DEBUG_OPTIMIZATION} Secundomer.Start; {$ENDIF}
  for i := 1 to N  do
    begin
    x := random(ClientWidth);
    y := random(ClientHeight);
    w := random(100) + 2*xR;
    h := random(100) + 2*xR;
    vRect := FixedRect(rect(x, y, x + w, y + h));
    //G32_Interface.gRectangleRounded(fBuffer, vRect , Fixed(xR), clTrBlue32 or $C0, fDrawningOptions);
    G32_Interface.gRectangleRR(fBuffer, FixedPoint(x, y) , Fixed(w), Fixed(h), Fixed(xR),
             vA, Color32(random(255), random(255), random(255),random(255)) , fDrawningOptions);
    end;
  {$IFDEF DEBUG_OPTIMIZATION} Secundomer.Stop; {$ENDIF}
  RepaintBuffer;
  {$IFDEF DEBUG_OPTIMIZATION} Secundomer.ShowResult; {$ENDIF}
end;


end.

